
/*
 *	SPM.cpp inlcudes the implementations for SPM.
 */


#include "SPM.h"

/*
 *	comp_fair_matching computes a fair assignment.
 *	
 *	(Memory statistics is not collected for this part)
 */
assignment* comp_fair_matching( data_t* data_v)
{
	configure *aConfigure;

	config_type *rtreeConf;
	rtree_info *rtreeInfo;

	int noOfOPoint, noOfPPoint;

	int m, M, dim, reinsert_p, no_histogram, assi_size;

	pointList *aOPointList;
	pointList *aPPointList;

	node_type *rootO;
	node_type *rootP;

	int i, j;

	assignment *aAssignment;

	int rtreeO_storage, rtreeP_storage, totalStorage;

	point *aPoint;

	twoData *aTwoData;

	triple *aTriple;

	aConfigure = configure_new();

	aOPointList = aConfigure->O;
	aPPointList = aConfigure->P;

	noOfOPoint = data_v->n_o_objs;
	noOfPPoint = data_v->n_p_objs;

	// R-tree Parameter Setting
	m = 25;
	M = 50;
	dim = data_v->dim;
	reinsert_p = 37;
	no_histogram = 12345; //dummy

	aConfigure->oFilename = NULL;
	aConfigure->pFilename = NULL;

	// transform the dataset from Cheng's format to Raymond's format

	// First, transform for O
	aOPointList->noOfPoint = noOfOPoint;
	aOPointList->noOfRemain = noOfOPoint;

	aOPointList->list = (point **) malloc(sizeof(point *)*noOfOPoint);


	for (i = 0; i < noOfOPoint; i++)
	{
		aPoint = point_new(dim);

		aPoint->id = i;
		// assume the weight is integer
		aPoint->weight = (int) data_v->o_objs[i].weight;
		for (j = 0; j < dim; j++)
		{
			aPoint->content[j] = data_v->o_objs[i].axis[j];
		}

		aOPointList->list[i] = aPoint;
	}

	// Second, transform for P
	aPPointList->noOfPoint = noOfPPoint;
	aPPointList->noOfRemain = noOfPPoint;

	aPPointList->list = (point **) malloc(sizeof(point *)*noOfPPoint);

	for (i = 0; i < noOfPPoint; i++)
	{
		aPoint = point_new(dim);

		aPoint->id = noOfOPoint + i;
		// assume the weight is integer
		aPoint->weight = (int) data_v->p_objs[i].weight;
		for (j = 0; j < dim; j++)
		{
			aPoint->content[j] = data_v->p_objs[i].axis[j];
		}

		aPPointList->list[i] = aPoint;
	}


	rtreeConf = aConfigure->rtreeConf;
	rtreeInfo = aConfigure->rtreeInfo;

	// put to rtreeConfig
	rtreeConf->m = m;
	rtreeConf->M = M;
	rtreeConf->dim = dim;
	rtreeConf->reinsert_p = reinsert_p;
	rtreeConf->no_histogram = no_histogram;
	strcpy(rtreeConf->positionfile, "");
	strcpy(rtreeConf->queryfile, "");

	// put to rtreeInfo
	rtreeInfo->m = m;
	rtreeInfo->M = M;
	rtreeInfo->dim = dim;
	rtreeInfo->reinsert_p = reinsert_p;
	rtreeInfo->no_histogram = no_histogram;
	rtreeInfo->extra_level = 0;

	//**** Finish the data transformation

	// Start the SPM algorithm

	globalStat= statistics_new();

	// cloning
	aTwoData = twoData_new();
	aTwoData->O = pointList_clone(aConfigure->O, dim);
	aTwoData->P = pointList_clone(aConfigure->P, dim);

#ifndef	WIN32
	struct rusage tree_sta, tree_end;
	get_cur_time( &tree_sta);
#endif

	// building r-tree
	printf("Building r-tree for O data..\n");
	rootO = quota_buildRtree(aOPointList, rtreeInfo);
	printf("Finished!\n");

	printf("Building r-tree for P data...\n");
	rootP = quota_buildRtree(aPPointList, rtreeInfo);
	printf("Finished!\n");

#ifndef	WIN32
	get_cur_time( &tree_end);

	float tmp;
	collect_time( &tree_sta, &tree_end, &swap_stat_v.R_tree_build_time, &tmp);
#endif

	// statistics
	statistics_rtreeNoOfNodeStorage(rootO, rootP, rtreeInfo);

	// calling the assignment algorithm
	aAssignment = spm_weightedChain(aTwoData, rootO, rootP, aConfigure);


	//***************** update Cheng's statistics
	//swap_stat_v.memory_v;
	//1. swap_stat_v.memory_max;
	rtreeO_storage = globalStat->rtreeO_storage;
	rtreeP_storage = globalStat->rtreeP_storage;

	assi_size = sizeof( assignment_s) + 
		sizeof( triple*) * aAssignment->noOfMatch + 
		sizeof( triple_s) * aAssignment->noOfMatch;
	totalStorage = rtreeO_storage + rtreeP_storage; 

	swap_stat_v.memory_v += totalStorage;
	if (swap_stat_v.memory_max < swap_stat_v.memory_v)
		swap_stat_v.memory_max = swap_stat_v.memory_v;

	//P point 2's axis 2: data_v->p_objs[1].axis[1];

	// FREE tree
	printf("Releasing r-tree for O data...\n");
	free_tree(rootO, rtreeInfo);
	printf("Releasing r-tree for P data...\n");
	free_tree(rootP, rtreeInfo);

	for (i = 0; i < aAssignment->noOfMatch; i++)
	{
		aTriple = aAssignment->list[i];

		aTriple->pid -= noOfOPoint;
	}
	
	return aAssignment;
}


/*
*	calc_fair_matching_mmd calculates the mmd value of a SPM matching.
*
* @data_v indicatest the input data for SPM matching problem.
*/
DOUBLE calc_fair_matching_mmd( data_t* data_v)
{
	int i;
	DOUBLE d_v, mmd = 0;

	//Performing the SPM matching.
	assignment* assi_v = comp_fair_matching( data_v);

	//Collecting the mmd value by comparing the matching distances.
	for( i=0; i<assi_v->noOfMatch; i++)
	{
		d_v = calc_dist( data_v->p_objs + assi_v->list[i]->pid, data_v->o_objs + assi_v->list[i]->oid, data_v->dim);
		if( d_v > mmd)
			mmd =  d_v;
	}

	//Releasing the resource related the assignment.
	for( i=0; i<assi_v->noOfMatch; i++)
		free( assi_v->list[i]);
	free( assi_v);

	return mmd;
}

/*
*	test_fair_matching runs the SPM algorithm.
*/
void test_fair_matching( )
{
	DOUBLE mmd;

	_config_t* cfg;
	data_t* data_v;

	FILE* o_fp, *s_fp;

	//Reading the configuration info.
	cfg = emp_read_config( );

	//Reading the raw data.
	data_v = emp_read_data( cfg);

	//Initializing the the statistic information.
	memset( &swap_stat_v, 0, sizeof( stat_t));


#ifndef WIN32
	struct rusage start_t, end_t;
	get_cur_time( &start_t);
#endif

	//Calculate the mmd value of the SPM matching.
	mmd = calc_fair_matching_mmd( data_v);

#ifndef WIN32
	float t_u, t_s;
	get_cur_time( &end_t);
	collect_time( &start_t, &t_rTree_end, &t_u, &t_s);

	//preprocessing time.
	swap_stat_v.preprocess_time = t_u;

	//matching time.
	collect_time( &t_rTree_end, &end_t, &t_u, &t_s);
	swap_stat_v.matching_time = t_u;
#endif

	//Printing the mmd value info & statistics info.
	if( (o_fp = fopen( SPM_RESULT_FILE, "w")) == NULL ||
		(s_fp = fopen( SWAP_STATISTIC_FILE, "w")) == NULL)
	{
		fprintf( stderr, "The result/statistic file cannot be opened.\n");
		exit(0);
	}

	fprintf( o_fp, "%lf\n", mmd);

	fprintf( s_fp, "%f\n%f\n", swap_stat_v.preprocess_time, swap_stat_v.matching_time);


	//Releasing the related resources.
	free( cfg);
	data_release( data_v);

	//fclose( c_fp);
	fclose( o_fp);
	fclose( s_fp);
}