/*
  This code was written by Raymond Chi-Wing Wong.
  Please read the following.

    Raymond Chi-Wing Wong, Yufei Tao, Ada Wai-Chee Fu and Xiaokui Xiao. 
    "On Efficient Spatial Matching", 
    the 33rd International Conference on Very Large Data Bases (VLDB'07), 
    Vienna, Austria on 23-28 Sept, 2007 

*/

#include "rsa.h"

extern void k_NN_search(node_type *root, R_TYPE *query, int k, NN_type **returnResult, rtree_info *aInfo);

void findOMatch(int oid, int pid, NN_type *oNode, twoData *aTwoData, node_type *rootO, node_type *rootP, configure *aConfigure, assignment *aAssignment, int chainLen)
{
	int k;
	R_TYPE *query;

	config_type *rtreeConf;
	rtree_info *rtreeInfo;

	int noOfDim;

	int isExist;
	NN_type *tempNN;
	NN_type *tempNN2;

	NN_type *Presult;

	point *oPoint;
	point *pPoint;

	triple *aTriple;
	node_type *aLeafNode;
	int inputLevel;

	int pid2;
	NN_type *pNode;

	int isEmpty;

	rtreeConf = aConfigure->rtreeConf;
	rtreeInfo = aConfigure->rtreeInfo;

	noOfDim = rtreeConf->dim;

	oPoint = twoData_getOPoint(aTwoData, oid);

	// set the query point o
	query = oPoint->content;

	// find NN of o
	k = 1;
	k_NN_search(rootP, query, k, &Presult, rtreeInfo);

	// statistics
	globalStat->totalNoOfNNQueryFromO++;

	// is p in NN of o?
	if (pid == -1)
	{
		isExist = 0;
	}
	else
	{
		isExist = 0;
		tempNN = Presult;
		while (tempNN != NULL)
		{
			if (tempNN->oid == pid)
			{
				isExist = 1;
				break;
			}

			tempNN = tempNN->next;
		}
	}

	// p is in NN of o
	if (isExist == 1)
	{
		// exist
		pPoint = twoData_getPPoint(aTwoData, pid);

		if (oPoint->weight > pPoint->weight)
		{
			// A <- A U {(o, p, q(p))}
			aTriple = triple_newValue(oid, pid, pPoint->weight);
			assignment_insert(aAssignment, &aTriple);

			// statistics
			statistics_update_chainLen(globalStat, chainLen);

			// decrement w(o) by q(p)
			oPoint->weight -= pPoint->weight;

			// set q(p) to 0 and remove p from P
			// remove from r-tree
			aLeafNode = tempNN->pointer;
			inputLevel = tempNN->level;
			condenseTree(rootP, aLeafNode, inputLevel, rtreeInfo);
			// remove from aTwoData
			pPoint->weight = 0;
			twoData_decrementP(aTwoData);	
			
			// statistics
			globalStat->totalNoOfDeletionFromP++;
		}
		else
		{
			// A <- A U {(o, p, w(o))}
			aTriple = triple_newValue(oid, pid, oPoint->weight);
			assignment_insert(aAssignment, &aTriple);

			// statistics
			statistics_update_chainLen(globalStat, chainLen);

			// decrement q(p) by w(o)
			pPoint->weight -= oPoint->weight;

			// set w(o) to 0 and remove o from O
			// remove from r-tree
			aLeafNode = oNode->pointer;
			inputLevel = oNode->level;
			condenseTree(rootO, aLeafNode, inputLevel, rtreeInfo);
			// remove from aTwoData
			oPoint->weight = 0;
			twoData_decrementO(aTwoData);

			// statistics
			globalStat->totalNoOfDeletionFromO++;

			if (pPoint->weight == 0)
			{
				// remove p from P
				aLeafNode = tempNN->pointer;
				inputLevel = tempNN->level;
				condenseTree(rootP, aLeafNode, inputLevel, rtreeInfo);
				twoData_decrementP(aTwoData);

				// statistics
				globalStat->totalNoOfDeletionFromP++;
			}

			return;
		}
	}

	// for each p' in P-{p}
	tempNN2 = Presult;
	while (tempNN2 != NULL)
	{
		pid2 = tempNN2->oid;
		if (pid2 != pid)
		{
			pNode = tempNN2;

			// findPMatch(p', o)
			findPMatch(pid2, oid, pNode, aTwoData, rootO, rootP, aConfigure, aAssignment, chainLen+1);

			// if o not in O 
			//    return
			if (oPoint->weight == 0)
			{
				return;
			}
		}

		tempNN2 = tempNN2->next;
	}

	// if P != emptyset
	//   findOMatch(o, p)
	isEmpty = twoData_pIsEmpty(aTwoData);
	if (isEmpty == 0)
	{
		findOMatch(oid, pid, oNode, aTwoData, rootO, rootP, aConfigure, aAssignment, chainLen);
	}

}

void findPMatch(int pid, int oid, NN_type *pNode, twoData *aTwoData, node_type *rootO, node_type *rootP, configure *aConfigure, assignment *aAssignment, int chainLen)
{
	int k;
	R_TYPE *query;

	config_type *rtreeConf;
	rtree_info *rtreeInfo;

	int noOfDim;

	int isExist;
	NN_type *tempNN;
	NN_type *tempNN2;

	NN_type *Oresult;

	point *oPoint;
	point *pPoint;

	triple *aTriple;
	node_type *aLeafNode;
	int inputLevel;

	int oid2;
	NN_type *oNode;

	int isEmpty;

	rtreeConf = aConfigure->rtreeConf;
	rtreeInfo = aConfigure->rtreeInfo;

	noOfDim = rtreeConf->dim;

	pPoint = twoData_getPPoint(aTwoData, pid);

	// set the query point p
	query = pPoint->content;

	// find NN of p
	k = 1;
	k_NN_search(rootO, query, k, &Oresult, rtreeInfo);

	// statistics
	globalStat->totalNoOfNNQueryFromP++;

	// is o in NN of p?
	if (oid == -1)
	{
		isExist = 0;
	}
	else
	{
		isExist = 0;
		tempNN = Oresult;
		while (tempNN != NULL)
		{
			if (tempNN->oid == oid)
			{
				isExist = 1;
				break;
			}

			tempNN = tempNN->next;
		}
	}

	// o is in NN of p
	if (isExist == 1)
	{
		// exist
		oPoint = twoData_getOPoint(aTwoData, oid);

		if (pPoint->weight > oPoint->weight)
		{
			// A <- A U {(o, p, w(o))}
			aTriple = triple_newValue(oid, pid, oPoint->weight);
			assignment_insert(aAssignment, &aTriple);

			// statistics
			statistics_update_chainLen(globalStat, chainLen);

			// decrement q(p) by w(o)
			pPoint->weight -= oPoint->weight;

			// set w(o) to 0 and remove o from O
			// remove from r-tree
			aLeafNode = tempNN->pointer;
			inputLevel = tempNN->level;
			condenseTree(rootO, aLeafNode, inputLevel, rtreeInfo);
			// remove from aTwoData
			oPoint->weight = 0;
			twoData_decrementO(aTwoData);	
			
			// statistics
			globalStat->totalNoOfDeletionFromO++;
		}
		else
		{
			// A <- A U {(o, p, q(p))}
			aTriple = triple_newValue(oid, pid, pPoint->weight);
			assignment_insert(aAssignment, &aTriple);

			// statistics
			statistics_update_chainLen(globalStat, chainLen);

			// decrement w(o) by q(p)
			oPoint->weight -= pPoint->weight;

			// set q(p) to 0 and remove p from P
			// remove from r-tree
			aLeafNode = pNode->pointer;
			inputLevel = pNode->level;
			condenseTree(rootP, aLeafNode, inputLevel, rtreeInfo);
			// remove from aTwoData
			pPoint->weight = 0;
			twoData_decrementP(aTwoData);

			// statistics
			globalStat->totalNoOfDeletionFromP++;

			if (oPoint->weight == 0)
			{
				// remove o from O
				aLeafNode = tempNN->pointer;
				inputLevel = tempNN->level;
				condenseTree(rootO, aLeafNode, inputLevel, rtreeInfo);
				twoData_decrementO(aTwoData);

				// statistics
				globalStat->totalNoOfDeletionFromO++;
			}

			return;
		}
	}

	// for each o' in O-{o}
	tempNN2 = Oresult;
	while (tempNN2 != NULL)
	{
		oid2 = tempNN2->oid;
		if (oid2 != oid)
		{
			oNode = tempNN2;

			// findOMatch(o', p)
			findOMatch(oid2, pid, oNode, aTwoData, rootO, rootP, aConfigure, aAssignment, chainLen+1);

			// if p not in P
			//    return
			if (pPoint->weight == 0)
			{
				return;
			}
		}

		tempNN2 = tempNN2->next;
	}

	// if O != emptyset
	//   findPMatch(p, o)
	isEmpty = twoData_oIsEmpty(aTwoData);
	if (isEmpty == 0)
	{
		findPMatch(pid, oid, pNode, aTwoData, rootO, rootP, aConfigure, aAssignment, chainLen);
	}

}

assignment *findStableAssignment(twoData *aTwoData, node_type *rootO, node_type *rootP, configure *aConfigure)
{
	assignment *returnValue;

	NN_type *oNode;
	int oid, pid;

	int isEmptyP, isEmptyO;

	int iterationNo;

	iterationNo = 0;

	returnValue = assignment_new();

	isEmptyP = twoData_pIsEmpty(aTwoData);
	isEmptyO = twoData_oIsEmpty(aTwoData);
	while ((isEmptyP == 0) && (isEmptyO == 0))
	{
		// randomly find a point o from O
		oNode = findRandomPoint(rootO, aConfigure->rtreeInfo);
		oid = oNode->oid;

		// create a non-existing point p
		pid = -1;

		// findOMatch(o, p)
		findOMatch(oid, pid, oNode, aTwoData, rootO, rootP, aConfigure, returnValue, 1);

		// remove p
		// do nothing

		// next iteration
		isEmptyP = twoData_pIsEmpty(aTwoData);
		isEmptyO = twoData_oIsEmpty(aTwoData);

		iterationNo++;
	}

	return returnValue;
}


void spm_findOMatch(int oid, NN_type *oNode, int *L, NN_type **nodeList, int *noOfEltL, twoData *aTwoData, node_type *rootO, node_type *rootP, configure *aConfigure, assignment *aAssignment)
{
	int k;
	R_TYPE *query;

	config_type *rtreeConf;
	rtree_info *rtreeInfo;

	int noOfDim;

	int isExist;
	NN_type *tempNN;
	NN_type *tempNN2;

	NN_type *Presult;

	point *oPoint;
	point *pPoint;

	triple *aTriple;
	node_type *aLeafNode;
	int inputLevel;

	int pid2;
	//NN_type *pNode;

	//int isEmpty;
	int pid;

	if ((*noOfEltL) <= 1)
	{
		pid = -1;
	}
	else
	{
		pid = L[(*noOfEltL)-2];
	}

	rtreeConf = aConfigure->rtreeConf;
	rtreeInfo = aConfigure->rtreeInfo;

	noOfDim = rtreeConf->dim;

	oPoint = twoData_getOPoint(aTwoData, oid);

	if (oPoint->weight == 0)
	{
		(*noOfEltL)--;
		return;
	}

	// set the queryt point o;
	query = oPoint->content;

	// find NN of o
	k = 1;
	k_NN_search(rootP, query, k, &Presult, rtreeInfo);

	// statistics
	globalStat->totalNoOfNNQueryFromO++;

	// is p in NN of o?
	if ((*noOfEltL) <= 1)
	{
		isExist = 0;
	}
	else
	{
		isExist = 0;
		tempNN = Presult;
		while (tempNN != NULL)
		{
			if (tempNN->oid == pid)
			{
				isExist = 1;
				break;
			}

			tempNN = tempNN->next;
		}
	}

	// p is in NN of o
	if (isExist == 1)
	{
		// exist
		pPoint = twoData_getPPoint(aTwoData, pid);

		if (oPoint->weight > pPoint->weight)
		{
			// A <- A U {(o, p, q(p))}
			aTriple = triple_newValue(oid, pid, pPoint->weight);
			assignment_insert(aAssignment, &aTriple);

			// statistics
			//statistics_update_chainLen(globalStat, chainLen);

			// decrement w(o) by q(p)
			oPoint->weight -= pPoint->weight;

			// set q(p) to 0 and remove p from P
			// remove from r-tree
			aLeafNode = tempNN->pointer;
			inputLevel = tempNN->level;
			condenseTree(rootP, aLeafNode, inputLevel, rtreeInfo);
			// remove from aTwoData
			pPoint->weight = 0;
			twoData_decrementP(aTwoData);	

			//*** remove o and p from L
			(*noOfEltL) -= 2;
			
			// statistics
			globalStat->totalNoOfDeletionFromP++;
		}
		else
		{
			// A <- A U {(o, p, w(o))}
			aTriple = triple_newValue(oid, pid, oPoint->weight);
			assignment_insert(aAssignment, &aTriple);

			// statistics
			//statistics_update_chainLen(globalStat, chainLen);

			// decrement q(p) by w(o)
			pPoint->weight -= oPoint->weight;

			// set w(o) to 0 and remove o from O
			// remove from r-tree
			aLeafNode = oNode->pointer;
			inputLevel = oNode->level;
			condenseTree(rootO, aLeafNode, inputLevel, rtreeInfo);
			// remove from aTwoData
			oPoint->weight = 0;
			twoData_decrementO(aTwoData);

			//*** remove o from L
			(*noOfEltL) -= 1;

			// statistics
			globalStat->totalNoOfDeletionFromO++;

			if (pPoint->weight == 0)
			{
				// remove p from P
				aLeafNode = tempNN->pointer;
				inputLevel = tempNN->level;
				condenseTree(rootP, aLeafNode, inputLevel, rtreeInfo);
				twoData_decrementP(aTwoData);

				//*** remove p from L
				(*noOfEltL) -= 1;

				// statistics
				globalStat->totalNoOfDeletionFromP++;
			}

			//return;
		}
	}

	// P' <- P - {p}
	// randomly choose an arbitrary p' in P'
	// insert p' into L

	// for each p' in P-{p}
	tempNN2 = Presult;
	while (tempNN2 != NULL)
	{
		pid2 = tempNN2->oid;
		if (pid2 != pid)
		{
			L[(*noOfEltL)] = pid2;
			nodeList[(*noOfEltL)] = tempNN2;
			(*noOfEltL)++;

			break;
		}

		tempNN2 = tempNN2->next;
	}
}


void spm_findPMatch(int pid, NN_type *pNode, int *L, NN_type **nodeList, int *noOfEltL, twoData *aTwoData, node_type *rootO, node_type *rootP, configure *aConfigure, assignment *aAssignment)
{
	int k;
	R_TYPE *query;

	int oid;

	config_type *rtreeConf;
	rtree_info *rtreeInfo;

	int noOfDim;

	int isExist;
	NN_type *tempNN;
	NN_type *tempNN2;

	NN_type *Oresult;

	point *oPoint;
	point *pPoint;

	triple *aTriple;
	node_type *aLeafNode;
	int inputLevel;

	int oid2;
	//NN_type *oNode;

	//int isEmpty;

	if ((*noOfEltL) <= 1)
	{
		oid = -1;
	}
	else
	{
		oid = L[(*noOfEltL)-2];
	}

	rtreeConf = aConfigure->rtreeConf;
	rtreeInfo = aConfigure->rtreeInfo;

	noOfDim = rtreeConf->dim;

	pPoint = twoData_getPPoint(aTwoData, pid);

	if (pPoint->weight == 0)
	{
		(*noOfEltL)--;
		return;
	}

	// set the query point p
	query = pPoint->content;

	// find NN of p
	k = 1;
	k_NN_search(rootO, query, k, &Oresult, rtreeInfo);

	// statistics
	globalStat->totalNoOfNNQueryFromP++;

	// is o in NN of p?
	if ((*noOfEltL) <= 1)
	{
		isExist = 0;
	}
	else
	{
		isExist = 0;
		tempNN = Oresult;
		while (tempNN != NULL)
		{
			if (tempNN->oid == oid)
			{
				isExist = 1;
				break;
			}

			tempNN = tempNN->next;
		}
	}

	// o is in NN of p
	if (isExist == 1)
	{
		// exist
		oPoint = twoData_getOPoint(aTwoData, oid);

		if (pPoint->weight > oPoint->weight)
		{
			// A <- A U {(o, p, w(o))}
			aTriple = triple_newValue(oid, pid, oPoint->weight);
			assignment_insert(aAssignment, &aTriple);

			// statistics
			//statistics_update_chainLen(globalStat, chainLen);

			// decrement q(p) by w(o)
			pPoint->weight -= oPoint->weight;

			// set w(o) to 0 and remove o from O
			// remove from r-tree
			aLeafNode = tempNN->pointer;
			inputLevel = tempNN->level;
			condenseTree(rootO, aLeafNode, inputLevel, rtreeInfo);
			// remove from aTwoData
			oPoint->weight = 0;
			twoData_decrementO(aTwoData);
			
			//*** remove o and p from L
			(*noOfEltL) -= 2;
			
			// statistics
			globalStat->totalNoOfDeletionFromO++;
		}
		else
		{
			// A <- A U {(o, p, q(p))}
			aTriple = triple_newValue(oid, pid, pPoint->weight);
			assignment_insert(aAssignment, &aTriple);

			// statistics
			//statistics_update_chainLen(globalStat, chainLen);

			// decrement w(o) by q(p)
			oPoint->weight -= pPoint->weight;

			// set q(p) to 0 and remove p from P
			// remove from r-tree
			aLeafNode = pNode->pointer;
			inputLevel = pNode->level;
			condenseTree(rootP, aLeafNode, inputLevel, rtreeInfo);
			// remove from aTwoData
			pPoint->weight = 0;
			twoData_decrementP(aTwoData);

			//*** remove p from L
			(*noOfEltL) -= 1;

			// statistics
			globalStat->totalNoOfDeletionFromP++;

			if (oPoint->weight == 0)
			{
				// remove o from O
				aLeafNode = tempNN->pointer;
				inputLevel = tempNN->level;
				condenseTree(rootO, aLeafNode, inputLevel, rtreeInfo);
				twoData_decrementO(aTwoData);

				// remove o from L
				(*noOfEltL) -= 1;

				// statistics
				globalStat->totalNoOfDeletionFromO++;
			}

			//return;
		}
	}

	// O' <- O - {o}
	// randomly choose an arbitrary o' in O'
	// insert o' into L

	// for each o' in O-{o}
	tempNN2 = Oresult;
	while (tempNN2 != NULL)
	{
		oid2 = tempNN2->oid;
		if (oid2 != oid)
		{
			L[(*noOfEltL)] = oid2;
			nodeList[(*noOfEltL)] = tempNN2;
			(*noOfEltL)++;

			break;
		}

		tempNN2 = tempNN2->next;
	}
}


assignment *spm_weightedChain(twoData *aTwoData, node_type *rootO, node_type *rootP, configure *aConfigure)
{
	assignment *returnValue;

	int *L;
	int noOfEltL;

	NN_type *oNode;
	NN_type *pNode;
	NN_type *aNode;

	NN_type **nodeList;

	int isEmptyP, isEmptyO;

	int noOfPointO;
	int noOfPointP;

	pointList *O;
	pointList *P;

	int oid, pid, id;

	int iterationNo, innerIterationNo;

	returnValue = assignment_new();

	O = aTwoData->O;
	P = aTwoData->P;

	noOfPointO = O->noOfPoint;
	noOfPointP = P->noOfPoint;

	L = (int *) malloc(sizeof(int)*(noOfPointO+noOfPointP)*2);
	memset(L, 0, sizeof(int)*(noOfPointO+noOfPointP)*2);

	nodeList = (NN_type **) malloc(sizeof(NN_type *)*(noOfPointO+noOfPointP)*2);
	memset(nodeList, 0, sizeof(NN_type *)*(noOfPointO+noOfPointP)*2);

	iterationNo = 0;

	isEmptyP = twoData_pIsEmpty(aTwoData);
	isEmptyO = twoData_oIsEmpty(aTwoData);
	while ((isEmptyP == 0) && (isEmptyO == 0))
	{
		// randomly finda point o from O
		oNode = findRandomPoint(rootO, aConfigure->rtreeInfo);
		oid = oNode->oid;

		// L <- {o}
		memset(L, 0, sizeof(int)*(noOfPointO+noOfPointP)*2);
		memset(nodeList, 0, sizeof(NN_type *)*(noOfPointO+noOfPointP)*2);
		L[0] = oid;
		nodeList[0] = oNode;
		noOfEltL = 1;

//if (iterationNo == 0)
//{
//	printf("");
//}

		innerIterationNo = 0;
		// while L != emptyset
		while ((noOfEltL > 0) && (isEmptyP == 0) && (isEmptyO == 0))
		{
			// x <-- last element of L
			id = L[noOfEltL-1];
			aNode = nodeList[noOfEltL-1];

//if (id < 0)
//{
//	printf("");
//}

//if (innerIterationNo == 12)
//{
//	printf("");
//}
//if (innerIterationNo == 3888)
//{
//	printf("");
//	{
//		FILE *filePtr;
//		int i;

//		filePtr = (FILE *) fopen("abc.txt", "w");
//		if (filePtr == NULL)
//		{
//			printf("File abc.txt cannot be written!\n");
//			exit(0);
//		}

//		for (i = 0; i < noOfEltL; i++)
//		{
//			fprintf(filePtr, "%d\n", L[i]);
//		}

//		fclose(filePtr);
//	}
//}
//if (L[noOfEltL-1] == 82)
//{
//	printf("");
//}
if ((iterationNo == 0) && (innerIterationNo == 19809))
{
	printf("");
}
			if (id < noOfPointO)
			{
				// x \in O
				oid = id;
				oNode = aNode;
				spm_findOMatch(oid, oNode, L, nodeList, &noOfEltL, aTwoData, rootO, rootP, aConfigure, returnValue);
			}
			else
			{
				// x \in P
				pid = id;
				pNode = aNode;
if (pid == 9073)
{
	printf("");
}
				spm_findPMatch(pid, pNode, L, nodeList, &noOfEltL, aTwoData, rootO, rootP, aConfigure, returnValue);
			}

			statistics_update_chainLen(globalStat, noOfEltL);

			innerIterationNo++;
			isEmptyP = twoData_pIsEmpty(aTwoData);
			isEmptyO = twoData_oIsEmpty(aTwoData);
		}

		iterationNo++;
		isEmptyP = twoData_pIsEmpty(aTwoData);
		isEmptyO = twoData_oIsEmpty(aTwoData);
	}

	free(L);
	free(nodeList);

	return returnValue;
}
